#pragma once

/// @brief Geometry data structures and algorithms
///
/// Tools for geometric data structures and operations, e.g. searching.
namespace dolfinx::geometry
{
}

// DOLFINx geometry interface

#include <dolfinx/geometry/BoundingBoxTree.h>
#include <dolfinx/geometry/gjk.h>
