#pragma once

/// @brief Miscellaneous classes, functions and types.
///
/// This namespace provides utility type functions for managing
/// subsystems, convenience classes and library-wide typedefs.
namespace dolfinx::common
{
}

// DOLFINx common

#include <dolfinx/common/MPI.h>
#include <dolfinx/common/Table.h>
#include <dolfinx/common/Timer.h>
#include <dolfinx/common/defines.h>
#include <dolfinx/common/timing.h>
#include <dolfinx/common/version.h>
